package br.usp.ime.coed.action;

import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;

import javax.swing.AbstractAction;
import javax.swing.AbstractButton;
import javax.swing.ImageIcon;
import javax.swing.JButton;
import javax.swing.JMenu;
import javax.swing.JMenuItem;
import javax.swing.JPopupMenu;
import javax.swing.JToolBar;
import javax.swing.border.BevelBorder;
import javax.swing.border.SoftBevelBorder;

/**
 * <p>This class permits that the same action be addicted to toolbars,
 *  menus. This facilitates a lot, since you can enable or disable
 *  this interface elements at once, sending the <code>setEnabled</code> message to
 *  the correspondent <code>FormattedAction</code> object.</p>
 *  Beside this, a <code>FormattedAction</code> object "formatts" itself when
 *  it?s being added to an interface. It results in a standard look,
 *  common to all the <code>FormattedAction</code> objects.
 *
 */

public abstract class FormattedAction extends AbstractAction {

  /** This is the mnemonic character of the action */
  private char mnemonic;

  /**
   * Contructor.
   * @param name is the name of the action. This value is used
   * in tool tip buttons and in menu items.
   * @param iconFileName this is the icon that represents the action.
   * It is be used in buttons and menu items.
   */
  public FormattedAction(final String name,String iconFileName){
    super(name, new ImageIcon(iconFileName));
  }

 /**
  * Constructor.
  * @param name is the name of the action. This value is used in tool
  * tip buttons and in menu items.
  * @param mnemonic the mnemonic character of the action
  * @param iconFileName this is the icon that represents the action.
  * It is used in buttons and in menu items.
  */
  public FormattedAction(final String name, char mnemonic, String iconFileName){
    super(name, new ImageIcon(iconFileName));
    this.mnemonic = mnemonic;
  }

  /**
   * This constructor generates an action without icon.
   * @param name is the name of the action. This value is used in tool
   * tip buttons and in menu items.
   * @param mnemonic the mnemonic character of the action
   */
  public FormattedAction(final String name, char mnemonic){
    super(name);
    this.mnemonic = mnemonic;
  }

  /**
   * Adds the <code>FormattedAction</code> to a <code>JToolBar</code>.
   * @param bar the <code>JToolBar</code> object
   * @return the <code>JButton</code> added to the <code>JToolBar</code>
   */
  public AbstractButton addTo(JToolBar bar){
    JButton button = bar.add(this);
    button.setToolTipText((String)getValue(NAME));
    setBorder(button);
    return button;
  }

  /**
   * Adds the <code>FormattedAction</code> to a <code>JMenu</code>.
   * @param menu the <code>JMenu</code> object
   * @return a <code>JMenuItem</code> item
   */
  public AbstractButton addTo(JMenu menu){

    JMenuItem menuItem = menu.add(this);
    menuItem.setMnemonic(mnemonic);
    return menuItem;
  }

  
  /**
   * Adds the <code>FormattedAction</code> to a <code>JPopUpMenu</code>.
   * @param menu the <code>JPopUpMenu</code> object
   * @return a <code>JMenuItem</code> item
   */
  public AbstractButton addTo(JPopupMenu menu){

  	JMenuItem menuItem = menu.add(this);
  	menuItem.setMnemonic(mnemonic);
  	return menuItem;
  }
  
  
  /**
   * This method does nothing. <br>It may be called by the process
   * that was initiated by <code>actionPerformed()</code> method to
   * indicate that it has finished.
   * <br>This should be overriden in special situations
   * (i.e. when it?s necessary to know that the process started
   * at the <code>actionPerformed</code> method call has finished)
   */
  public void actionFinalized(){ }

 /**
   * Sets the border of the especified button.
   * @param button a <code>JButton</code> object
   */
  protected void setBorder(final JButton button){

    button.setBorder(new SoftBevelBorder(BevelBorder.RAISED));
    button.setBorderPainted(false);
    button.addMouseListener(new MouseAdapter(){
      public void mouseEntered(MouseEvent e){
        button.setBorderPainted(true);
      }
      public void mouseExited(MouseEvent e){
        button.setBorderPainted(false);}
      }
    );
  }

  /**
   * Contructor. Should not be used.
   */
  public FormattedAction(){super();}

}
