/*
 * Criado em 24/10/2004
 */
package br.usp.ime.coed.model;

import java.util.LinkedList;
import java.util.Observable;

/**
 * Controls the redo and undo operations
 * 
 * An instance of this class is an observable object. The observer are instances
 * of:
 * 
 * @see br.usp.ime.coed.action.edit.RedoAction
 * @see br.usp.ime.coed.action.edit.UndoAction
 * 
 * For more information. see the Memento pattern at 
 * "Design Pattern" (Gang of four)
 * 
 * @see br.usp.ime.model.CoedDocument
 * @see br.usp.ime.model.CoedDocumentMemento
 * 
 * @author Andr Yoshiaki Kashiwabara <yoshiaki@vision.ime.usp.br>
 */
public class CoedDocumentCareTaker extends Observable {

    /**
     * List of <code>CoedDocumentMemento</code> for undo operation
     */
    private LinkedList undos;

    /**
     * List of <code>CoedDocumentMemento</code> for redo operation
     */
    private LinkedList redos;

    /**
     * When the state of this class object changes, we have to notify it to the
     * observers
     */
    public static final String CHANGED = "CARE_TAKER_CHANGED";

    /**
     * Creates the undos and redos list
     */
    public CoedDocumentCareTaker() {
        undos = new LinkedList();
        redos = new LinkedList();
    }

    /**
     * Add a CoedMemento to undos list and clear the redo list, notify the
     * changes to observers
     * 
     * @param m
     */
    public void addMemento(CoedDocumentMemento m) {
        undos.add(m);
        redos.clear();
        setChanged();
        notifyObservers(CHANGED);
    }

    /**
     * The undo operation
     * 
     * @param doc
     *            the current document state
     * @return a CoedDocumentMemento
     */
    public CoedDocumentMemento undo(CoedDocument doc) {
        if (undos.size() <= 0)
            return null;
        CoedDocumentMemento last = (CoedDocumentMemento) undos.removeLast();
        redos.add(doc.createMemento());
        doc.setNotSaved(true);
        setChanged();
        notifyObservers(CHANGED);
        return (CoedDocumentMemento) last;
    }

    /**
     * The redo operation
     * 
     * @param doc
     *            the current document state
     * @return
     */
    public CoedDocumentMemento redo(CoedDocument doc) {
        if (redos.size() <= 0)
            return null;
        CoedDocumentMemento last = (CoedDocumentMemento) redos.removeLast();
        undos.add(doc.createMemento());
        doc.setNotSaved(true);
        setChanged();
        notifyObservers(CHANGED);
        return (CoedDocumentMemento) last;
    }

    /**
     * @return Returns the redo list size
     */
    public int redoSize() {
        return redos.size();
    }

    /**
     * @return Retuins the undo list size
     */
    public int undoSize() {
        return undos.size();
    }

    /**
     * notify observer that "this" object is active
     */
    public void notifyActive() {
        setChanged();
        notifyObservers(CHANGED);
    }

    /**
     * clear the undos e redos.
     */
    public void clear() {
        undos.clear();
        redos.clear();
        setChanged();
        notifyObservers(CHANGED);
    }
}