/*
 * Created on 22/10/2004
 */
package br.usp.ime.coed.model;

import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.io.LineNumberReader;
import java.util.HashMap;
import java.util.Observable;

/**
 * Responsible for managing all the <i>open, save, close, etc </i> operations
 * 
 * An instance of this class is an observable object. The observer is instance
 * of:
 * 
 * @see br.usp.ime.coed.view.TabbedPane
 * 
 * @see br.usp.ime.model.CoedDocument
 * @see br.usp.ime.model.CoedDocumentMemento
 * @see br.usp.ime.model.CoedDocumentCareTaker
 * 
 * @author Andr Yoshiaki Kashiwabara <yoshiaki@vision.ime.usp.br>
 */

public class CoedDocumentManager extends Observable {
    /**
     * All opened documents. Maps the name of the document to the document
     */
    HashMap documents;

    /**
     * The current document --> document that is on focus
     */
    private CoedDocument currentDocument;

    /**
     * Each document must be associated to a DocumentCareTaker
     */
    private HashMap documentsToDocumentCareTaker;

    /**
     * The current document was changed
     */
    public static final String CURRENT_CHANGED = "DOC_MANAGER_CURRENT_CHANGED";

    /**
     * The current document was saved
     */
    public static final String SAVE = "DOC_MANAGER_SAVE";

    /**
     * A new document was created
     */
    public static final String NEW = "DOC_MANAGER_NEW";

    /**
     * A document was opened
     */
    public static final String OPEN = "DOC_MANAGER_OPEN";

    /**
     * A document was closed
     */
    public static final String CLOSE = "DOC_MANAGER_CLOSE";

    /**
     * Inicialize the CoedDocumentManager
     *  
     */
    public CoedDocumentManager() {
        documents = new HashMap();
        documentsToDocumentCareTaker = new HashMap();
    }

    /**
     * Close the current document
     */
    public void closeDocument() {
        if(currentDocument == null)
            return;
        documents.remove(currentDocument.getName());
        CoedDocumentCareTaker careTaker = (CoedDocumentCareTaker) getDocumentCareTaker(currentDocument);
        careTaker.clear();
        documentsToDocumentCareTaker.remove(currentDocument);
        setChanged();
        notifyObservers(CLOSE);
    }

    /**
     * Creates a new document
     * 
     * @param path
     *            is the document parent path
     * @param name
     *            is the document name
     * @throws IllegalStateException
     *             A new document has the same name of an opened document
     */
    public void newDocument(String path, String name)
            throws IllegalStateException {
        if (!documents.containsKey(name)) {
            CoedDocument doc = new CoedDocument(name);
            doc.setPath(path);
            CoedDocumentCareTaker caretaker = new CoedDocumentCareTaker();
            documents.put(name, doc);
            documentsToDocumentCareTaker.put(doc, caretaker);
            currentDocument = doc;
            setChanged();
            notifyObservers(NEW);
        } else {
            throw new IllegalStateException("The document " + name
                    + " is already opened !");
        }
    }

    /**
     * Opens the document
     * 
     * @param path
     *            parent path
     * @param name
     *            the document name
     * @throws Exception -
     *             the document is alread opened
     */
    public void openDocument(String path, String name) throws Exception {
        StringBuffer xml = new StringBuffer();
        File inFile = new File(path + "/" + name);
        LineNumberReader input = new LineNumberReader(new FileReader(inFile));
        String line;
        while ((line = input.readLine()) != null)
            xml.append(line).append("\r\n");
        input.close();
        if (!documents.containsKey(name)) {
            CoedDocumentMemento memento = new CoedDocumentMemento();
            memento.setState(xml.toString());
            CoedDocument open = new CoedDocument();
            documents.put(name, open);
            documentsToDocumentCareTaker.put(open, new CoedDocumentCareTaker());
            open.setMemento(memento);
            open.setName(name);
            open.setPath(path);
            currentDocument = open;
            setChanged();

            notifyObservers(OPEN);
        } else {
            throw new IllegalStateException("The document " + name
                    + " is already opened !");
        }
    }

    /**
     * Import a .cnf document
     * 
     * @param path
     *            the parent path
     * @param name
     *            the file name
     * @throws Exception --
     *             the document with the same name is already opened
     */
    public void openEGeneDocument(String path, String name) throws Exception {
        StringBuffer egeneString = new StringBuffer();
        File inFile = new File(path + "/" + name);
        LineNumberReader input = new LineNumberReader(new FileReader(inFile));
        String line;
        while ((line = input.readLine()) != null)
            egeneString.append(line).append("\r\n");
        input.close();
        name = name + ".gen";
        if (!documents.containsKey(name)) {
            CoedDocumentMemento memento = new CoedDocumentMemento();
            memento.setStateFromEGeneFile(egeneString.toString());
            CoedDocument open = new CoedDocument();
            documents.put(name, open);
            documentsToDocumentCareTaker.put(open, new CoedDocumentCareTaker());
            open.setMemento(memento);
            open.setName(name);
            open.setPath(path);
            currentDocument = open;
            setChanged();
            notifyObservers(OPEN);
        } else {
            throw new IllegalStateException("The document " + name
                    + " is already opened !");
        }
    }

    /**
     * Implements the save as operation
     * 
     * @param path
     *            the parent path
     * @param name
     *            the name of the file
     * @throws Exception
     */
    public void saveAsDocument(String path, String name) throws Exception {

        String xml = currentDocument.createMemento().xmlString();
        BufferedOutputStream out = new BufferedOutputStream(
                new FileOutputStream(path + "/" + name));
        out.write(xml.getBytes());
        out.close();
        documents.remove(currentDocument.getName());
        currentDocument.setName(name);
        currentDocument.setPath(path);
        currentDocument.setNotSaved(false);
        documents.put(currentDocument.getName(), currentDocument);
        setChanged();
        notifyObservers(SAVE);
    }

    /**
     * Save the current document
     * @throws Exception 
     */
    public void saveDocument() throws Exception {
        String xml = currentDocument.createMemento().xmlString();
        BufferedOutputStream out = new BufferedOutputStream(
                new FileOutputStream(currentDocument.getPath() + "/"
                        + currentDocument.getName()));
        out.write(xml.getBytes());
        out.close();
        currentDocument.setNotSaved(false);
        setChanged();
        notifyObservers(SAVE);
    }

    /**
     * @return Returns the currentDocument.
     */
    public CoedDocument getCurrentDocument() {
        return currentDocument;
    }

    /**
     * Set the current document and notify the changes to the observer
     * @param currentDocument
     *            The currentDocument to set.
     */
    public void setCurrentDocument(CoedDocument currentDocument) {
        this.currentDocument = currentDocument;
        setChanged();
        notifyObservers(CURRENT_CHANGED);
    }

    /**
     * Get the DocumentCareTaker associated with a document
     * @param doc - the document
     * @return <code>CoedDocumentCareTaker</code>
     */
    public CoedDocumentCareTaker getDocumentCareTaker(CoedDocument doc) {
        return (CoedDocumentCareTaker) documentsToDocumentCareTaker.get(doc);
    }

    /**
     * Returns true if the file is opened
     * @param file the file name
     * @return <code>boolean</code>
     */
    public boolean containsDocument(String file) {
        return documents.containsKey(file);
    }

    /**
     * Get the Map of documents
     * @return <code>HashMap</code>
     */
    public HashMap getDocuments() {
        return documents;
    }

    /**
     *  Updates the observers
     */
    public void updateObserver() {
        setChanged();
        notifyObservers(CURRENT_CHANGED);
    }

    /**
     * Export the document to EGene configuration format
     * @param path
     * @param file
     */
    public void saveAsEGeneDocument(String path, String file) throws Exception {
        String egene = currentDocument.toString();
        BufferedOutputStream out = new BufferedOutputStream(
                new FileOutputStream(path + "/" + file));
        out.write(egene.getBytes());
        out.close();

    }

    /**
     * Save the current document state
     * 
     * @see CoedDocumentCareTaker.
     */
    public void saveCurrentDocumentState() {
        CoedDocument document = getCurrentDocument();
        CoedDocumentCareTaker caretaker = getDocumentCareTaker(document);
        caretaker.addMemento(document.createMemento());
    }
}

