package br.usp.ime.coed.view;


import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Point;
import java.awt.Stroke;
import java.awt.geom.Rectangle2D;


/**
 * The graphical representation of a pipe.
 * <br>Is constituted by a line and an arrow.
 */
public class Pipe {

  /** The number of pixels that an arrow ocupies */
  private static final int delta = 15;

  /** The color of the pipe label */
  private Color labelColor = Color.black;

  /** The color of the main line */
  private Color lineColor;

  /** The width of the pipe arrow */
  private int arrowWidth;

  /** The length of the pipe arrow */
  private int arrowLength;

  /** Indicates if the pipe is dashed */
  private boolean dashed;

  /** Auxiliary point (improvement in performance) */
  private Point pf, pt, p1, p2;

  /**  One of the <code>Pipe</code> extremes */
  private GraphicalElement from, to;

  /** The pipe line stroke */
  private Stroke stroke;

  /** The pipe name (if any) */
  private String labelName;
  
  /**
   * Constructor.
   * @param from the beggining pipe element
   * @param to the end pipe element
   * @param pipeName the pipe name
   * @param lineColor the pipe line color
   * @param arrowWidth the pipe arrow width
   * @param arrowLength the pipe arrow length
   */
  public Pipe(GraphicalElement from, GraphicalElement to, String pipeName,
              Color lineColor, int arrowWidth, int arrowLength){

    this(from, to, pipeName, lineColor, false, arrowWidth, arrowLength);
  }

  /**
   * Constructor.
   * @param from the beggining pipe element
   * @param to the end pipe element
   * @param format the pipe format
   */
  public Pipe(GraphicalElement from, GraphicalElement to, PipeFormat format){
      this(from, to, "", format.getLineColor(), format.isDashed(),
      format.getArrowWidth(), format.getArrowLength());
    }

  /**
   * Constructor.
   * @param from the beggining pipe element
   * @param to the end pipe element
   * @param pipeName the pipe name
   * @param lineColor the pipe line color
   * @param dashed indicates if the pipe line is dashed or not
   * @param arrowWidth the pipe arrow width
   * @param arrowLength the pipe arrow length
   */
  public Pipe( GraphicalElement from, GraphicalElement to, String pipeName,
                  Color lineColor, boolean dashed, int arrowWidth, int arrowLength){

    this.lineColor    = lineColor;
    this.arrowLength  = arrowLength;
    this.arrowWidth   = arrowWidth;
    this.from = from;
    this.to = to;
    this.labelName = pipeName;
    this.dashed = dashed;

    pf = new Point();
    pt = new Point();
    p1 = new Point();
    p2 = new Point();

    if(dashed)
      stroke = new BasicStroke(2.0F, BasicStroke.CAP_ROUND, BasicStroke.JOIN_BEVEL,
                               10.0F, new float[]{ 8, 4}, 0 );
    else
      stroke =  new BasicStroke(2.0F, BasicStroke.CAP_ROUND, BasicStroke.JOIN_BEVEL );
  }

  /**
   * Paints the pipe on <code>g</code>.
   * @param g the <code>Graphics2d</code> where the paint is made on.
   */
  void paintOn( Graphics2D g){

    Rectangle2D fromR = from.getBounds(),toR   = to.getBounds();

    /* Calculate auxiliary points */
    if(fromR.getMaxX() < toR.getMinX() - delta)//toR.getCenterX() - 10)
      calculatePoints(pf, pt, from, to);
    else if(fromR.getMinX() > toR.getMaxX()+delta)//toR.getMaxX()+10)
      calculatePoints(pt, pf, to, from);
    else if(fromR.getMaxX() < toR.getCenterX() - delta)
      calculateHPartiallyJoined(pf, pt, from, to);
    else if(fromR.getMinX() > toR.getCenterX() + delta)
      calculateHPartiallyJoined(pt, pf, to, from);
    else{
      if(fromR.getMaxY() < toR.getMinY() - delta)
        calculateHorizontallyJoined(pt, pf, to, from);
      else if(fromR.getMinY() > toR.getMaxY() + delta)
        calculateHorizontallyJoined(pf, pt, from, to);
      else if(fromR.getMaxY() < toR.getCenterY() - delta)
        calculateVPartiallyJoined(pt, pf, to, from);
      else if(fromR.getMinY() > toR.getCenterY() + delta)
        calculateVPartiallyJoined(pf, pt, from, to);
      else calculateJoined(pf, pt, fromR, toR);
    }

    /*Draw lines */
    g.setColor( getLineColor() );
    Stroke oldStroke = g.getStroke();
    g.setStroke( stroke );
	if(p2.x == pt.x && pt.y > p2.y){
		pt.y -= 20;
		p2.y -= 10;
		p1.y = p2.y;
	} 
	if(p1.x == pf.x && pf.y > p1.y){
		pf.y -= 20;
		p1.y -= 10;
		p2.y = p1.y;
	}
    g.drawLine( pf.x, pf.y, p1.x, p1.y );
    g.drawLine( p1.x, p1.y, p2.x, p2.y );
    g.drawLine( p2.x, p2.y, pt.x, pt.y );
    // g.drawLine( pf.x, pf.y, pt.x, pt.y );

    /* Draw arrow */
    drawArrow( g, p2, pt );
    // drawArrow(g, pf, pt);
    g.setStroke( oldStroke );

    /* Draw name, if defined */
    if(labelName != null)
      drawLabel( g );
  }

  /**
   * Auxiliary points calculation method (called when the pipe extremes
   * are horizontally and vertically superposed).
   * @param j1 a auxiliary point (p1 or p2)
   * @param j2 the other auxiliary point (p1 or p2)
   * @param b1 the bounds of an extreme element
   * @param b2 the bounds of the other extreme element
   */
  private void calculateJoined(Point j1, Point j2, Rectangle2D b1, Rectangle2D b2){

    int deltaX, deltaY;

    if(!from.lineAtSide(GraphicalElement.LEFT) &&
       !to.lineAtSide(GraphicalElement.LEFT)){

      j1.x = minX(b1);     j2.x = minX(b2);
      deltaX = -20;        deltaY = 0;

      if(b1.getMinY() < b2.getMinY()){
        j1.y = minY(b1) + 3;   j2.y = maxY(b2) - 3;
      }
      else{
        j1.y = maxY(b1) - 3;   j2.y = minY(b2) + 3;
      }

      from.setLineAtSide(GraphicalElement.LEFT);
      to.setLineAtSide(GraphicalElement.LEFT);
    }

    else if(!from.lineAtSide(GraphicalElement.RIGHT) &&
            !to.lineAtSide(GraphicalElement.RIGHT)){
      j1.x = maxX(b1);   j2.x = maxX(b2);
      deltaX = 20;       deltaY = 0;

      if(b1.getMinY() < b2.getMinY()){
         j1.y = minY(b1) + 3;   j2.y = maxY(b2) - 3;
      }
      else{
        j1.y = maxY(b1) - 3;   j2.y = minY(b2) + 3;
      }

      from.setLineAtSide(GraphicalElement.RIGHT);
      to.setLineAtSide(GraphicalElement.RIGHT);
    }

    else if(!from.lineAtSide(GraphicalElement.TOP) &&
            !to.lineAtSide(GraphicalElement.TOP)){
      j1.y = minY(b1);   j2.y = minY(b2);
      deltaX = 0;        deltaY = -20;

      if(b1.getMinX() < b2.getMinX()){
        j1.x = minX(b1) + 3;   j2.x = maxX(b2) - 3;
      }
      else{
        j1.x = maxX(b1) - 3;   j2.x = minX(b2) + 3;
      }

      from.setLineAtSide(GraphicalElement.TOP);
      to.setLineAtSide(GraphicalElement.TOP);
    }

    else{
      j1.y = maxY(b1);   j2.y = maxY(b2);
      deltaX = 0;        deltaY = 20;

      if(b1.getMinX() < b2.getMinX()){
         j1.x = minX(b1) + 3;   j2.x = maxX(b2) - 3;
      }
      else{
        j1.x = maxX(b1) - 3;   j2.x = minX(b2) + 3;
      }

      from.setLineAtSide(GraphicalElement.BOTTOM);
      to.setLineAtSide(GraphicalElement.BOTTOM);
    }

    p1.x = pf.x + deltaX;  p2.x = pt.x + deltaX;
    p1.y = pf.y + deltaY;  p2.y = pt.y + deltaY;

  }// calculateJoined

  /**
   * Auxiliary points calculation method (called when the pipe extremes
   * are horizontally superposed).
   * @param lower one of the auxiliary points (pf or pt)
   * @param upper the other auxiliary point (pf or pt)
   * @param lowerG the lower extreme element
   * @param upperG the upper extreme element
   */
  private void calculateHorizontallyJoined(Point lower, Point upper,
                  GraphicalElement lowerG,GraphicalElement upperG){

    Rectangle2D lowerR = lowerG.getBounds(), upperR = upperG.getBounds();

    if(!lowerG.lineAtSide(GraphicalElement.TOP)&&
       !upperG.lineAtSide(GraphicalElement.BOTTOM)){

      lower.x = halfX(lowerR);      lower.y = minY(lowerR);
      upper.x = halfX(upperR);      upper.y = maxY(upperR);
      p1.x = pf.x;                  p2.x = pt.x;
      p2.y = p1.y = (int) (lowerR.getMinY() - (lowerR.getMinY() - upperR.getMaxY())/2);

      lowerG.setLineAtSide(GraphicalElement.TOP);
      upperG.setLineAtSide(GraphicalElement.BOTTOM);
    }

    /** @todo: testar se sai para fora do painel */
    else if(!lowerG.lineAtSide(GraphicalElement.RIGHT)&&
            !upperG.lineAtSide(GraphicalElement.RIGHT)){

      lower.x = maxX(lowerR);       lower.y = halfY(lowerR);
      upper.x = maxX(upperR);       upper.y = halfY(upperR);
      p1.y = pf.y;                  p2.y = pt.y;
      p1.x = p2.x = Math.max(lower.x, upper.x) + delta;

      lowerG.setLineAtSide(GraphicalElement.RIGHT);
      upperG.setLineAtSide(GraphicalElement.RIGHT);
    }

    else if(!lowerG.lineAtSide(GraphicalElement.LEFT)&&
            !upperG.lineAtSide(GraphicalElement.LEFT)){

      lower.x = minX(lowerR);       lower.y = halfY(lowerR);
      upper.x = minX(upperR);       upper.y = halfY(upperR);
      p1.y = pf.y;                  p2.y = pt.y;
      p1.x = p2.x = Math.min(lower.x, upper.x) - delta;

      lowerG.setLineAtSide(GraphicalElement.LEFT);
      upperG.setLineAtSide(GraphicalElement.LEFT);
    }
  }// calculateHorizontallyJoined

  /**
   * Auxiliary points calculation method (called when the pipe extremes
   * are partially horizontally superposed).
   * @param left one of the auxiliary points (pf or pt)
   * @param right the other auxiliary point (pf or pt)
   * @param leftG the left extreme element
   * @param rightG the right extreme element
   */
  private void calculateHPartiallyJoined(Point left, Point right,
                  GraphicalElement leftG, GraphicalElement rightG){

    boolean drawn = false;
    Rectangle2D leftR = leftG.getBounds(), rightR = rightG.getBounds();

    if(leftR.getMaxY() < rightR.getCenterY()-delta){
      if(!leftG.lineAtSide(GraphicalElement.BOTTOM)&&
         !rightG.lineAtSide(GraphicalElement.LEFT)){

        //Nesse caso, fao um L utilizando as posies testadas
        left.x = halfX(leftR);      left.y = maxY(leftR);
        right.x = minX(rightR);     right.y = halfY(rightR);
        p2.x = p1.x = left.x;       p2.y = p1.y = right.y;

        leftG.setLineAtSide(GraphicalElement.BOTTOM);
        rightG.setLineAtSide(GraphicalElement.LEFT);
        drawn = true;
      }
      else if(!leftG.lineAtSide(GraphicalElement.RIGHT)&&
              !rightG.lineAtSide(GraphicalElement.TOP)){

        left.x = maxX(leftR);       left.y = halfY(leftR);
        right.x = halfX(rightR);    right.y = minY(rightR);
        p1.x = p2.x = right.x;      p1.y = p2.y = left.y;

        leftG.setLineAtSide(GraphicalElement.RIGHT);
        rightG.setLineAtSide(GraphicalElement.TOP);
        drawn = true;
      }

      else if(!leftG.lineAtSide(GraphicalElement.RIGHT)&&
              !rightG.lineAtSide(GraphicalElement.RIGHT)){

        left.x = maxX(leftR);       left.y = halfY(leftR);
        right.x = maxX(rightR);     right.y = minY(rightR);
        p1.y = pf.y;                p2.y = pt.y;
        p1.x = p2.x = right.x + 20;

        leftG.setLineAtSide(GraphicalElement.RIGHT);
        rightG.setLineAtSide(GraphicalElement.RIGHT);
        drawn = true;
      }

      else if(!leftG.lineAtSide(GraphicalElement.LEFT) &&
              !rightG.lineAtSide(GraphicalElement.LEFT)){

        left.x = minX(leftR);      left.y = halfY(leftR);
        right.x = minX(rightR);    right.y = halfY(rightR);
        p1.y = pf.y;                p2.y = pt.y;
        p1.x = p2.x = left.x - 20;

        leftG.lineAtSide(GraphicalElement.LEFT);
        rightG.lineAtSide(GraphicalElement.LEFT);
        drawn = true;
      }
    }//if

    else if(leftR.getMinY() > rightR.getCenterY()+delta){
      if(!leftG.lineAtSide(GraphicalElement.TOP)&&
         !rightG.lineAtSide(GraphicalElement.LEFT)){

        left.x = halfX(leftR);        left.y = minY(leftR);
        right.x = minX(rightR);       right.y = halfY(rightR);
        p1.x = p2.x = left.x;         p1.y = p2.y = right.y;

        leftG.setLineAtSide(GraphicalElement.TOP);
        rightG.setLineAtSide(GraphicalElement.LEFT);
        drawn = true;
      }
      else if(!leftG.lineAtSide(GraphicalElement.RIGHT)&&
              !rightG.lineAtSide(GraphicalElement.BOTTOM)){

        left.x = maxX(leftR);         left.y = halfY(leftR);
        right.x = halfX(rightR);      right.y = maxY(rightR);
        p1.x = p2.x = right.x;        p1.y = p2.y = left.y;

        leftG.setLineAtSide(GraphicalElement.RIGHT);
        rightG.setLineAtSide(GraphicalElement.BOTTOM);
        drawn = true;
      }
      else if(!leftG.lineAtSide(GraphicalElement.RIGHT)&&
              !rightG.lineAtSide(GraphicalElement.RIGHT)){

        left.x = maxX(leftR);         left.y = halfY(leftR);
        right.x = maxX(rightR);       right.y = halfY(rightR);
        p1.y = pf.y;                  p2.y = pt.y;
        p1.x = p2.x =  right.x + 20;

        leftG.setLineAtSide(GraphicalElement.RIGHT);
        rightG.setLineAtSide(GraphicalElement.LEFT);
        drawn = true;
      }
      else if(!leftG.lineAtSide(GraphicalElement.LEFT)&&
              !rightG.lineAtSide(GraphicalElement.LEFT)){

        left.x = minX(leftR);        left.y = halfY(leftR);
        right.x = minX(rightR);      right.y = halfY(rightR);
        p1.x = pf.x;                  p2.x = pt.x;
        p1.y = p2.y = left.x -20;

        leftG.setLineAtSide(GraphicalElement.TOP);
        rightG.setLineAtSide(GraphicalElement.BOTTOM);
        drawn = true;
      }
    }//else if

    if(!drawn){
      if(leftR.getMinY() <= rightR.getMinY()-delta){
        if(!leftG.lineAtSide(GraphicalElement.BOTTOM) &&
           !rightG.lineAtSide(GraphicalElement.LEFT)){

          left.x = halfX(leftR);       left.y = maxY(leftR);
          right.x = minX(rightR);      right.y = maxY(rightR)- 5;
          p1.x = p2.x = left.x;        p1.y = p2.y = right.y;

          leftG.setLineAtSide(GraphicalElement.BOTTOM);
          rightG.setLineAtSide(GraphicalElement.LEFT);
          drawn = true;
        }
        else if(!leftG.lineAtSide(GraphicalElement.RIGHT) &&
                !rightG.lineAtSide(GraphicalElement.TOP)){

          left.x = maxX(leftR);       left.y = minY(leftR) + 5;
          right.x = halfX(rightR);      right.y = minY(rightR);
          p1.x = p2.x = right.x;        p1.y = p2.y = left.y;

          leftG.setLineAtSide(GraphicalElement.RIGHT);
          rightG.setLineAtSide(GraphicalElement.TOP);
          drawn = true;
        }
      }
      else if(rightR.getMinY() <= leftR.getMinY()-delta){
        if(!leftG.lineAtSide(GraphicalElement.TOP) &&
           !rightG.lineAtSide(GraphicalElement.LEFT)){

          left.x = halfX(leftR);       left.y = minY(leftR);
          right.x = minX(rightR);      right.y = minY(rightR)+ 5;
          p1.x = p2.x = left.x;        p1.y = p2.y = right.y;

          leftG.setLineAtSide(GraphicalElement.TOP);
          rightG.setLineAtSide(GraphicalElement.LEFT);
          drawn = true;
        }
        else if(!leftG.lineAtSide(GraphicalElement.RIGHT) &&
                !rightG.lineAtSide(GraphicalElement.BOTTOM)){

          left.x = maxX(leftR);       left.y = maxY(leftR) - 5;
          right.x = halfX(rightR);      right.y = maxY(rightR);
          p1.x = p2.x = right.x;        p1.y = p2.y = left.y;

          leftG.setLineAtSide(GraphicalElement.RIGHT);
          rightG.setLineAtSide(GraphicalElement.BOTTOM);
          drawn = true;
        }
      }

      if(drawn == false){
        if(!leftG.lineAtSide(GraphicalElement.TOP) &&
           !rightG.lineAtSide(GraphicalElement.TOP)){

          left.x = halfX(leftR);       left.y = minY(leftR);
          right.x = halfX(rightR);     right.y = minY(rightR);
          p1.x = pf.x;                 p2.x = pt.x;
          p1.y = p2.y = left.y - 20;

          leftG.setLineAtSide(GraphicalElement.TOP);
          rightG.setLineAtSide(GraphicalElement.TOP);
        }
        else if(!leftG.lineAtSide(GraphicalElement.BOTTOM) &&
                !rightG.lineAtSide(GraphicalElement.BOTTOM)){

          left.x = halfX(leftR);       left.y = maxY(leftR);
          right.x = halfX(rightR);     right.y = maxY(rightR);
          p1.x = pf.x;                 p2.x = pt.x;
          p1.y = p2.y = right.y + 20;

          leftG.setLineAtSide(GraphicalElement.BOTTOM);
          rightG.setLineAtSide(GraphicalElement.BOTTOM);
        }
        else
          calculateJoined(left, right, leftR, rightR);
      }
    }//else

  }// calculateHPartiallyJoined

  /**
   * Auxiliary points calculation method (called when the pipe extremes
   * are partially vertically superposed).
   * @param lower one of the auxiliary points (pf or pt)
   * @param upper the other auxiliary point (pf or pt)
   * @param lowerG the lower extreme element
   * @param upperG the upper extreme element
   */
  private void calculateVPartiallyJoined(Point lower, Point upper,
                  GraphicalElement lowerG,GraphicalElement upperG){

    boolean drawn = false;
    Rectangle2D lowerR = lowerG.getBounds(), upperR = upperG.getBounds();

    if(lowerR.getMinX() <= upperR.getMinX() - delta){
      if(!lowerG.lineAtSide(GraphicalElement.TOP) &&
         !upperG.lineAtSide(GraphicalElement.LEFT)){

        lower.x = minX(lowerR) + 5;   lower.y = minY(lowerR);
        upper.x = minX(upperR);      upper.y = halfY(upperR);
        p1.x = p2.x = lower.x;       p1.y = p2.y = upper.y;

        lowerG.setLineAtSide(GraphicalElement.TOP);
        upperG.setLineAtSide(GraphicalElement.LEFT);
        drawn = true;
      }
      else if(!lowerG.lineAtSide(GraphicalElement.RIGHT) &&
              !upperG.lineAtSide(GraphicalElement.BOTTOM)){

        lower.x = maxX(lowerR);       lower.y = halfY(lowerR);
        upper.x = maxX(upperR) - 5;   upper.y = maxY(upperR);
        p1.x = p2.x = upper.x;        p1.y = p2.y = lower.y;

        lowerG.setLineAtSide(GraphicalElement.RIGHT);
        upperG.setLineAtSide(GraphicalElement.BOTTOM);
        drawn = true;
      }
    }//if
    else if(upperR.getMinX() <= lowerR.getMinX()-delta){
      if(!lowerG.lineAtSide(GraphicalElement.TOP) &&
         !upperG.lineAtSide(GraphicalElement.RIGHT)){

        lower.x = maxX(lowerR) - 5;  lower.y = minY(lowerR);
        upper.x = maxX(upperR);      upper.y = halfY(upperR);
        p1.x = p2.x = lower.x;        p1.y = p2.y = upper.y;

        lowerG.setLineAtSide(GraphicalElement.TOP);
        upperG.setLineAtSide(GraphicalElement.RIGHT);
        drawn = true;
      }
      else if(!lowerG.lineAtSide(GraphicalElement.LEFT) &&
              !upperG.lineAtSide(GraphicalElement.BOTTOM)){

        lower.x = minX(lowerR);      lower.y = halfY(lowerR);
        upper.x = minX(upperR) + 5; upper.y = maxY(upperR);
        p1.x = p2.x = upper.x;        p1.y = p2.y = lower.y;

        lowerG.setLineAtSide(GraphicalElement.RIGHT);
        upperG.setLineAtSide(GraphicalElement.BOTTOM);
        drawn = true;
      }
    }//else if

    if(drawn == false){
      if(!lowerG.lineAtSide(GraphicalElement.RIGHT) &&
         !upperG.lineAtSide(GraphicalElement.RIGHT)){

        lower.x = maxX(lowerR);      lower.y = halfY(lowerR);
        upper.x = maxX(upperR);      upper.y = halfY(upperR);
        p1.y = pf.y;                 p2.y = pt.y;
        p1.x = p2.x = Math.max(lower.x, upper.x) + 20;

        lowerG.setLineAtSide(GraphicalElement.RIGHT);
        upperG.setLineAtSide(GraphicalElement.RIGHT);
      }
      else if(!lowerG.lineAtSide(GraphicalElement.LEFT) &&
              !upperG.lineAtSide(GraphicalElement.LEFT)){

        lower.x = minX(lowerR);      lower.y = halfY(lowerR);
        upper.x = minX(upperR);      upper.y = halfY(upperR);
        p1.y = pf.y;                 p2.y = pt.y;
        p1.x = p2.x = Math.min(lower.x, upper.x) - 20;

        lowerG.setLineAtSide(GraphicalElement.LEFT);
        upperG.setLineAtSide(GraphicalElement.LEFT);
      }
      else
        calculateJoined(lower, upper, lowerR, upperR);
    }//if
  }// calculateVPartiallyJoined

  /**
   * Auxiliary points calculation method (called when the pipe extremes
   * are not superposed).
   * @param left one of the auxiliary points (pf or pt)
   * @param right the other auxiliary point (pf or pt)
   * @param leftG the left extreme element
   * @param rightG the right extreme element
   */
  private void calculatePoints(Point left, Point right, GraphicalElement leftG,
				                      GraphicalElement rightG){
      
      Rectangle2D leftR = leftG.getBounds(), rightR = rightG.getBounds();
      
      if(leftR.getMaxY() < rightR.getMinY()-delta){
	  if(!leftG.lineAtSide(GraphicalElement.TOP) &&
	     !rightG.lineAtSide(GraphicalElement.BOTTOM)) {
	      
	      left.x = halfX(leftR);      left.y = maxY(leftR);
	      right.x = halfX(rightR);   right.y = minY(rightR);
	      p1.x = pf.x;                p2.x = pt.x;
	      p1.y = p2.y = left.y + (right.y - left.y)/2;
	      
	      leftG.lineAtSide(GraphicalElement.TOP);
	      rightG.lineAtSide(GraphicalElement.BOTTOM);
	  }
	  else if(!leftG.lineAtSide(GraphicalElement.BOTTOM)&&
		  !rightG.lineAtSide(GraphicalElement.LEFT)){
	      
	      //Nesse caso, fao um L utilizando as posies testadas
	      left.x = halfX(leftR);      left.y = maxY(leftR);
	      right.x = minX(rightR);     right.y = halfY(rightR);
	      p2.x = p1.x = left.x;       p2.y = p1.y = right.y;
	      
	      leftG.setLineAtSide(GraphicalElement.BOTTOM);
	      rightG.setLineAtSide(GraphicalElement.LEFT);
	  }
	  else if(!leftG.lineAtSide(GraphicalElement.RIGHT)&&
		  !rightG.lineAtSide(GraphicalElement.TOP)){
	      
	      left.x = maxX(leftR);       left.y = halfY(leftR);
	      right.x = halfX(rightR);    right.y = minY(rightR);
	      p1.x = p2.x = right.x;      p1.y = p2.y = left.y;
	      
	      leftG.setLineAtSide(GraphicalElement.RIGHT);
	      rightG.setLineAtSide(GraphicalElement.TOP);
	  } 
	  else { /* if(!leftG.lineAtSide(GraphicalElement.RIGHT)&&
		  !rightG.lineAtSide(GraphicalElement.LEFT)) */
	      
	      left.x = maxX(leftR);       left.y = halfY(leftR);
	      right.x = minX(rightR);     right.y = halfY(rightR);
	      p1.y = pf.y;                p2.y = pt.y;
	      p1.x = p2.x = (int)(Math.abs(right.x - left.x)/2);

	      leftG.setLineAtSide(GraphicalElement.RIGHT);
	      rightG.setLineAtSide(GraphicalElement.LEFT);
	  
	  }//if
      }
      else if(leftR.getMinY()/**/ > rightR.getMaxY()+delta){
	  if (!leftG.lineAtSide(GraphicalElement.TOP)&&
	      !rightG.lineAtSide(GraphicalElement.BOTTOM)) {
	      
	      left.x = halfX(leftR);        left.y = minY(leftR);
	      right.x = halfX(rightR);      right.y = maxY(rightR);
	      p1.x = pf.x;                  p2.x = pt.x;
	      p1.y = p2.y = Math.abs(right.y - left.y)/2 + Math.min(left.y, right.y);
	      
	      leftG.setLineAtSide(GraphicalElement.TOP);
	      rightG.setLineAtSide(GraphicalElement.BOTTOM);
	  }else if(!leftG.lineAtSide(GraphicalElement.TOP)&&
	     !rightG.lineAtSide(GraphicalElement.LEFT)){
	      
	      left.x = halfX(leftR);        left.y = minY(leftR);
	      right.x = minX(rightR);       right.y = halfY(rightR);
	      p1.x = p2.x = left.x;         p1.y = p2.y = right.y;
	      
	      leftG.setLineAtSide(GraphicalElement.TOP);
	      rightG.setLineAtSide(GraphicalElement.LEFT);
	  }
	  else if(!leftG.lineAtSide(GraphicalElement.RIGHT)&&
		  !rightG.lineAtSide(GraphicalElement.BOTTOM)){
	      
	      left.x = maxX(leftR);         left.y = halfY(leftR);
	      right.x = halfX(rightR);      right.y = maxY(rightR);
	      p1.x = p2.x = right.x;        p1.y = p2.y = left.y;
	      
	      leftG.setLineAtSide(GraphicalElement.RIGHT);
	      rightG.setLineAtSide(GraphicalElement.BOTTOM);
	  }
	  else { /*if(!leftG.lineAtSide(GraphicalElement.RIGHT)&&
		   !rightG.lineAtSide(GraphicalElement.LEFT)){*/
	      
	      left.x = maxX(leftR);         left.y = halfY(leftR);
	      right.x = minX(rightR);       right.y = halfY(rightR);
	      p1.y = pf.y;                  p2.y = pt.y;
	      p1.x = p2.x =  left.x + Math.abs(right.x - left.x)/2;
	      
	      leftG.setLineAtSide(GraphicalElement.RIGHT);
	      rightG.setLineAtSide(GraphicalElement.LEFT);
	  }
      }//else if
      else{
	  if(!leftG.lineAtSide(GraphicalElement.RIGHT)&&
	     !rightG.lineAtSide(GraphicalElement.LEFT)){
	      
	      left.x = maxX(leftR);     left.y = halfY(leftR);
	      right.x = minX(rightR);   right.y = halfY(rightR);
	      p1.y = pf.y;              p2.y = pt.y;
	      p1.x = p2.x = left.x + (right.x - left.x)/2;
	      
	      leftG.setLineAtSide(GraphicalElement.RIGHT);
	      rightG.setLineAtSide(GraphicalElement.LEFT);
	  }
	  //testar se vai sair fora da janela o desenho da seta
	  else if(!leftG.lineAtSide(GraphicalElement.TOP)&&
		  !rightG.lineAtSide(GraphicalElement.TOP)){
	      
	      left.x = halfX(leftR);    left.y = minY(leftR);
	      right.x = halfX(rightR);  right.y = minY(rightR);
	      p1.x = pf.x;              p2.x = pt.x;
	      p1.y = p2.y = Math.min(left.y, right.y) - delta;
	      
	      leftG.setLineAtSide(GraphicalElement.TOP);
	      rightG.setLineAtSide(GraphicalElement.TOP);
	  }
	  else{// if(!leftG.lineAtSide(GraphicalElement.BOTTOM)&&
	      //    !rightG.lineAtSide(GraphicalElement.BOTTOM)){
	      
	      left.x = halfX(leftR);    left.y = maxY(leftR);
	      right.x = halfX(rightR);   right.y = maxY(rightR);
	      p1.x = pf.x;              p2.x = pt.x;
	      p1.y = p2.y = Math.max(left.y, right.y) + delta;
	      
	      leftG.setLineAtSide(GraphicalElement.BOTTOM);
	      rightG.setLineAtSide(GraphicalElement.BOTTOM);
	  }
      }//else
  } // calculatePoints
    
    /**
     * Draws an arrow at the head of the line pointing from the <CODE>from</CODE>
     * pointto the <CODE>to</CODE> point, using the arrow length, arrow width and
     * closed option as specified in the constructor.
     *
     * If to.equals( from ), nothing will be draw.
     * @param g the <code>Graphics2D</code>
     * @param from the <code>from</code> element
     * @param to the <code>to</code>
     */
    private final void drawArrow(Graphics2D g, Point from, Point to){
	drawArrow( g, from, to, this.arrowLength, this.arrowWidth);
    }
    
    /**
     * Draws an arrow at the head of the line pointing from the <CODE>from</CODE>
     * point to the <CODE>to</CODE> point, using <CODE>a</CODE> as arrow length and
     * <CODE>b</CODE> as arrow width.
     *
     *If to.equals( from ), nothing will be draw.
     *
     * @param g the <code>Graphics2D</code>
     * @param from the <code>from</code> element
     * @param to the <code>to</code>
     *@param a the arrow length
     *@param b the arrow width
     */
    private final void drawArrow( Graphics2D g, Point from, Point to, int a, int b){

	final float deltaX = to.x - from.x;
	final float deltaY = to.y - from.y;
	final float modulus = (float)Math.sqrt( deltaX * deltaX + deltaY * deltaY );

	// Check for zero.
	if (modulus != 0){

	    final float sinAlpha = deltaY / modulus;
	    final float cosAlpha = deltaX / modulus;

	    final int tx = (int) (to.x - a * cosAlpha);
	    final int ty = (int) (to.y - a * sinAlpha);

	    final int rx = (int) (tx + b * sinAlpha);
	    final int ry = (int) (ty - b * cosAlpha);
	    final int sx = (int) (tx - b * sinAlpha);
	    final int sy = (int) (ty + b * cosAlpha);

	    // Draw.
	    int[] xPoints = { rx, to.x, sx };
	    int[] yPoints = { ry, to.y, sy };

	    g.fillPolygon( xPoints, yPoints, 3);
	}
    }

    /**
     * Draws the pipe name (if defined) in the middle point of from and to.
     *
     * @param r
     * @param from
     * @param to
     */
    void drawLabel(Graphics g){

	if ( labelName != null ){

	    g.setColor( this.labelColor );
	    g.drawString( labelName,
			  Math.abs( new Double((from.getX() + to.getX()) / 2 - 20).intValue()),
			  Math.abs( new Double((from.getY() + to.getY()) / 2).intValue()));
	} // if
    }

    /**
     * Returns the label color.
     * @return a <code>Color</code> object
     */
    private Color getLabelColor(){
	return labelColor;
    }

    /**
     * @return the color of the line for the Relationships to be drawn.
     */
    final Color getLineColor(){
	return this.lineColor;
    }

    /**
     * Returns the from drawing element extreme.
     * @return a <code>GraphicalElement</code>
     */
    public final GraphicalElement getFromDE(){
	return from;
    }

    /**
     * Returns the to drawing element extreme.
     * @return a <code>GraphicalElement</code>
     */
    public final GraphicalElement getToDE(){
	return to;
    }

    /**
     * Returns the center x of r.
     * @param r a <code>Rectangle2D</code> bound
     * @return an int value
     */
    private final int halfX(Rectangle2D r){
	return (int)(r.getMinX() + r.getWidth()/2);
    }

    /**
     * Returns the center y of r.
     * @param r a <code>Rectangle2D</code> bound
     * @return an int value
     */
    private final int halfY(Rectangle2D r){
	return (int)(r.getMinY() + r.getHeight()/2);
    }

    /**
     * Returns the minimum x of r.
     * @param r a <code>Rectangle2D</code> bound
     * @return an int value
     */
    private final int minX(Rectangle2D r){
	return (int)r.getMinX();
    }

    /**
     * Returns the minimum y of r.
     * @param r a <code>Rectangle2D</code> bound
     * @return an int value
     */
    private final int minY(Rectangle2D r){
	return (int)r.getMinY();
    }

    /**
     * Returns the maximum x of r.
     * @param r a <code>Rectangle2D</code> bound
     * @return an int value
     */
    private final int maxX(Rectangle2D r){
	return (int)r.getMaxX();
    }

    /**
     * Returns the maximum y of r.
     * @param r a <code>Rectangle2D</code> bound
     * @return an int value
     */
    private final int maxY(Rectangle2D r){
	return (int)r.getMaxY();
    }

}// Pipe

